@@ -1,5 +1,5 @@
 // winclip.c (part of mintty)
-// Copyright 2008-12 Andy Koppe, 2018 Thomas Wolff
+// Copyright 2008-23 Andy Koppe, 2018 Thomas Wolff
 // Adapted from code from PuTTY-0.60 by Simon Tatham and team.
 // Licensed under the terms of the GNU General Public License v3 or later.
 
@@ -19,6 +19,46 @@
 #include <sys/cygwin.h>  // cygwin_internal
 #endif
 
+
+#ifdef check_cygdrive
+// Adapt /cygdrive prefix to actual configured one (like / or anything else);
+// this experimental approach is not enabled as it is only used for WSL and
+// dynamic adaptation is hardly useful.
+
+static char * _cygdrive = 0;
+static wchar * _wcygdrive = 0;
+
+static char *
+cygdrive(void)
+{
+  if (!_cygdrive) {
+    char target [99];
+    int ret = readlink ("/proc/cygdrive", target, sizeof (target) - 1);
+    if (ret >= 0) {
+      target [ret] = '\0';
+      _cygdrive = strdup(target);
+    }
+    else
+#if defined(__MSYS__) || defined(__MINGW32)
+      _cygdrive = "/";
+#else
+      _cygdrive = "/cygdrive";
+#endif
+    _wcygdrive = cs__mbstowcs(target);
+  }
+  return _cygdrive;
+}
+
+static wchar *
+wcygdrive(void)
+{
+  (void)cygdrive();
+  return _wcygdrive;
+}
+
+#endif
+
+
 static DWORD WINAPI
 shell_exec_thread(void *data)
 {
@@ -29,7 +69,8 @@ shell_exec_thread(void *data)
   cygwin_internal(CW_SYNC_WINENV);
 #endif
 
-  if ((INT_PTR)ShellExecuteW(wnd, 0, wpath, 0, 0, SW_SHOWNORMAL) <= 32) {
+  SetLastError(ERROR_PATH_NOT_FOUND);  // in case !*wpath
+  if (!*wpath || (INT_PTR)ShellExecuteW(wnd, 0, wpath, 0, 0, SW_SHOWNORMAL) <= 32) {
     uint error = GetLastError();
     if (error != ERROR_CANCELLED) {
       int msglen = 1024;
@@ -57,7 +98,7 @@ shell_exec(wstring wpath)
   // this frees wpath via shell_exec_thread
 }
 
-#define dont_debug_wsl
+#define dont_debug_wslpath 3
 
 static struct {
   string p;
@@ -86,22 +127,6 @@ ispathprefix(string pref, string path)
   return false;
 }
 
-static bool
-ispathprefixw(wstring pref, wstring path)
-{
-  if (*pref == '/')
-    pref++;
-  if (*path == '/')
-    path++;
-  int len = wcslen(pref);
-  if (0 == wcsncmp(pref, path, len)) {
-    path += len;
-    if (!*path || *path == '/')
-      return true;
-  }
-  return false;
-}
-
 static char * wsl_conf_mnt = 0;
 static struct {
   char * dev_fs;
@@ -117,6 +142,97 @@ skip(char * s)
   return s;
 }
 
+static int
+pathpref(char * x, char * y)
+{
+  char _x = *x;
+  char _y = *y;
+  if (!_x) {
+    if (!_y)
+      return 2;  // equal
+    else
+      return 1;  // prefix
+  }
+  if (_x == '\\')
+    _x = '/';
+  if (_y == '\\')
+    _y = '/';
+  if (_x != _y)
+    return 0;
+  return pathpref(x + 1, y + 1);
+}
+
+#if 0
+static char *
+strsubst(char * str, char c, char d)
+{
+  char * s = str;
+  while ((s = strchr(s, c)))
+    *s = d;
+  return str;
+}
+#endif
+
+#ifdef debug_wslpath
+static inline void
+show_fstab(void)
+{
+#if defined(debug_wslpath) && debug_wslpath > 2
+  for (int i = 0; i < wsl_fstab_len; i++)
+    printf("wslpath[%d] <%s> <%s>\n", i, wsl_fstab[i].dev_fs, wsl_fstab[i].mount_point);
+#endif
+}
+#endif
+
+static void
+windrives(void)
+{
+  HANDLE hnet;
+  if (WNetOpenEnum(RESOURCE_CONNECTED, RESOURCETYPE_DISK, 0, NULL, &hnet))
+    return;
+
+  DWORD bufsiz = 16384;
+  NETRESOURCEW * nr = malloc(bufsiz);
+  for (;;) {
+    //int res = WNetEnumResourceW(hnet, &(DWORD){1}, nr, &bufsiz);
+    DWORD nrn = 1;
+    int res = WNetEnumResourceW(hnet, &nrn, nr, &bufsiz);
+    if (res == NO_ERROR || res == ERROR_MORE_DATA) {
+#if defined(debug_wslpath) && debug_wslpath > 1
+      printf("net use <%ls> <%ls>\n", nr->lpLocalName, nr->lpRemoteName);
+#endif
+      if (!nr->lpLocalName)
+        continue;  // skip false matches on UNC path net mounts
+
+      // we accept some initial memory leak of configuration strings here
+      char * localdrive = cs__wcstoutf(nr->lpLocalName);
+      char * remoteshare = cs__wcstoutf(nr->lpRemoteName);
+      // adjust fstab with {"X:", "\\host\share"}
+      for (int i = 0; i < wsl_fstab_len; i++) {
+        // for {"X:", "\\host\share"},
+        // and all wsl_fstab entries 
+        // {"//host/share", "/mnt/..."} or
+        // {"\\host\share", "/mnt/..."},
+        // replace "//host/share" with "X:",
+        // also with subdir entries like
+        // {"//host/share/subdir", "/mnt/..."}
+        // replace "//host/share/subdir" with "X:\subdir"
+        switch (pathpref(remoteshare, wsl_fstab[i].dev_fs)) {
+          when 2: wsl_fstab[i].dev_fs = localdrive;
+          when 1: {  // prefix
+                  char * subdir = wsl_fstab[i].dev_fs + strlen(remoteshare);
+                  wsl_fstab[i].dev_fs = asform("%s%s", localdrive, subdir);
+          }
+        }
+      }
+    }
+    else
+      break;
+  }
+
+  WNetCloseEnum(hnet);
+}
+
 static bool
 wslmntmapped(void)
 {
@@ -174,6 +290,9 @@ wslmntmapped(void)
           if (x-- > p2 && *x == '/')
             *x = 0;
           if (*p1 && *p2) {
+#if defined(debug_wslpath)
+            printf("fstab <%s> <%s>\n", p1, p2);
+#endif
             wsl_fstab = renewn(wsl_fstab, wsl_fstab_len + 1);
             wsl_fstab[wsl_fstab_len].dev_fs = strdup(p1);
             wsl_fstab[wsl_fstab_len].mount_point = strdup(p2);
@@ -188,10 +307,91 @@ wslmntmapped(void)
     free(wsl$conf);
     free(rootfs);
     wslmnt_init = true;
+
+    // substitute fstab dev_fs entries with Windows notation
+    windrives();
   }
   return wsl_conf_mnt || wsl_fstab;
 }
 
+static char *
+wslpath(char * path)
+{
+  char * res = null;
+  if (wslmntmapped() &&
+#if defined(__MSYS__) || defined(__MINGW32)
+      path[2] == '/'
+#else
+      ispathprefix("/cygdrive", path)
+#endif
+     )
+  {  // check drive path
+#if defined(__MSYS__) || defined(__MINGW32)
+#else
+    path += 9;
+#endif
+    if (*path == '/')
+      path++;  // point to drive letter
+    if (wsl_conf_mnt) {  // [automount] configured
+      res = asform("%s/%s", wsl_conf_mnt, path);
+    }
+    else if (path[1] == '/') {
+      for (int i = 0; i < wsl_fstab_len; i++) {
+#if defined(debug_wslpath) && debug_wslpath > 2
+        printf("wslpath <%s> fstab <%s> <%s>\n", path, wsl_fstab[i].dev_fs, wsl_fstab[i].mount_point);
+#endif
+        if (tolower((int)*path) == tolower((int)wsl_fstab[i].dev_fs[0])
+         && wsl_fstab[i].dev_fs[1] == ':'
+           )
+        {
+          char * dev = &wsl_fstab[i].dev_fs[2];
+          if (strlen(dev) == 1) {
+            //path <x/dir/file> fstab <X:\> </mnt/mountpoint>
+            res = asform("%s%s", wsl_fstab[i].mount_point, path + 1);
+            break;
+          }
+          else if (pathpref(dev, path + 1)) {
+            //path <x/dir/file> fstab <X:\subdir> </mnt/mountpoint>
+            res = asform("%s%s", wsl_fstab[i].mount_point, path + 1 + strlen(dev));
+            break;
+          }
+        }
+      }
+    }
+  }
+  else if (wslmntmapped()) {  // check UNC path
+    for (int i = 0; i < wsl_fstab_len; i++) {
+#if defined(debug_wslpath) && debug_wslpath > 2
+      printf("wslpath <%s> <%s> (%s)\n", path, wsl_fstab[i].mount_point, wsl_fstab[i].dev_fs);
+#endif
+      if (pathpref(wsl_fstab[i].dev_fs, path)) {
+        res = asform("%s%s", wsl_fstab[i].mount_point, path + strlen(wsl_fstab[i].dev_fs));
+        break;
+      }
+    }
+  }
+  // cygwin-internal paths not supported
+  return res;
+}
+
+#ifdef pathname_conversion_here
+
+static bool
+ispathprefixw(wstring pref, wstring path)
+{
+  if (*pref == '/')
+    pref++;
+  if (*path == '/')
+    path++;
+  int len = wcslen(pref);
+  if (0 == wcsncmp(pref, path, len)) {
+    path += len;
+    if (!*path || *path == '/')
+      return true;
+  }
+  return false;
+}
+
 static char *
 unwslpath(wchar * wpath)
 {
@@ -201,46 +401,42 @@ unwslpath(wchar * wpath)
     if (wsl_conf_mnt && ispathprefix(wsl_conf_mnt, wslpath))
       res = asform("/cygdrive%s", &wslpath[strlen(wsl_conf_mnt)]);
     else
-      for (int i = 0; i < wsl_fstab_len; i++)
-        if (ispathprefix(wsl_fstab[i].mount_point, wslpath))
+      for (int i = 0; i < wsl_fstab_len; i++) {
+#if defined(debug_wslpath) && debug_wslpath > 2
+        printf("unwslpath <%s> <%s> (%s)\n", wslpath, wsl_fstab[i].mount_point, wsl_fstab[i].dev_fs);
+#endif
+        if (ispathprefix(wsl_fstab[i].mount_point, wslpath)) {
           if (wsl_fstab[i].dev_fs[1] == ':') {
-            res = asform("/cygdrive/%c%s", 
-                         tolower(wsl_fstab[i].dev_fs[0]), 
+            res = asform(
+#if defined(__MSYS__) || defined(__MINGW32)
+#else
+                         "/cygdrive"
+#endif
+                         "/%c%s%s", 
+                         tolower((int)wsl_fstab[i].dev_fs[0]), 
+                         &wsl_fstab[i].dev_fs[2], 
                          &wslpath[strlen(wsl_fstab[i].mount_point)]);
             break;
           }
+          else {  // handle UNC path mount
+            res = asform("%s%s", wsl_fstab[i].dev_fs, wslpath + strlen(wsl_fstab[i].mount_point));
+            for (char * r = res; *r; r++)
+              // cygwin path conversion fails on backslash UNC paths
+              if (*r == '\\')
+                *r = '/';
+          }
+        }
+      }
     free(wslpath);
   }
   return res;
 }
 
-static char *
-wslpath(char * path)
-{
-  char * res = null;
-  if (ispathprefix("/cygdrive", path) && wslmntmapped()) {
-    path += 10;  // point to drive letter
-    if (wsl_conf_mnt)
-      res = asform("%s/%s", wsl_conf_mnt, path);
-    else if (path[1] == '/')
-      for (int i = 0; i < wsl_fstab_len; i++)
-        if (tolower(*path) == tolower(wsl_fstab[i].dev_fs[0])
-         && wsl_fstab[i].dev_fs[1] == ':'
-         && wsl_fstab[i].dev_fs[2] == 0) {
-            path ++;
-            res = asform("%s%s", wsl_fstab[i].mount_point, path);
-            break;
-          }
-  }
-  // cygwin-internal paths not supported
-  return res;
-}
-
 wchar *
 dewsl(wchar * wpath)
 {
-#ifdef debug_wsl
-  printf("open <%ls>\n", wpath);
+#ifdef debug_wslpath
+  printf("dewsl <%ls>\n", wpath);
 #endif
   char * unwslp = unwslpath(wpath);
   if (unwslp) {
@@ -302,65 +498,95 @@ dewsl(wchar * wpath)
   return wpath;
 }
 
+#endif
+
+#if CYGWIN_VERSION_API_MINOR < 206
+#define wcsncasecmp wcsncmp
+#define wmemcpy(t, s, n)	memcpy(t, s, 2 * n)
+#endif
+
 void
 win_open(wstring wpath, bool adjust_dir)
 // frees wpath
 {
-  // unescape
-  int wl = wcslen(wpath);
-  if ((*wpath == '"' || *wpath == '\'') && wpath[wl - 1] == *wpath) {
-    // don't wpath++ (later delete!)
-    wchar * p0 = (wchar *)wpath;
-    for (int i = 0; i < wl - 2; i++)
-      p0[i] = p0[i + 1];
-    p0[wl - 2] = 0;
+  size_t wl = wcslen(wpath);
+  wchar *wbuf = newn(wchar, wl + 1);
+
+  if (wl > 2 && (*wpath == '"' || *wpath == '\'') && wpath[wl - 1] == *wpath) {
+    // Remove pair of leading and trailing quotes.
+    wl -= 2;
+    wmemcpy(wbuf, wpath + 1, wl);
+  }
+  else if (!wcsncmp(wpath, W("\\\\"), 2)) {
+    // Assume that a path starting with two backslashes is a Windows UNC path
+    // and copy it unmodified.
+    wmemcpy(wbuf, wpath, wl);
   }
   else {
-    wchar * p0 = (wchar *)wpath;
-    wchar * p1 = p0;
-    while (*p0) {
-      if (*p0 == '\\')
-        p0++;
-      if (*p0)
-        *p1++ = *p0++;
+    // Remove backslashes, but only if they precede shell special characters.
+    // Other backslashes are more likely to be Windows path separators.
+    wstring p = wpath;
+    wl = 0;
+    while (*p) {
+      if (*p == '\\' && wcschr(W(" \t\n|&;<>()$`\\\"'*?![]#~=%^"), p[1]))
+        p++;
+      wbuf[wl++] = *p++;
     }
-    *p1 = 0;
+  }
+  wbuf[wl] = 0;
+  //printf("win_open %ls wbuf %ls\n", wpath, wbuf);
+  delete(wpath);
+
+  // check for URL
+  wchar *p = wbuf;
+  while (iswalpha(*p)) p++;
+  if (*p == ':' && p - wbuf > 2) {
+    shell_exec(wbuf); // frees wbuf
+    return;
   }
 
-  wstring p = wpath;
-  while (iswalpha(*p)) p++;
-  if (*wpath == '\\' || *p == ':' || wcsncmp(W("www."), wpath, 4) == 0) {
-    // Looks like it's a Windows path or URI
-    shell_exec(wpath); // frees wpath
-  }
-  else {
+  // guard file opening against foreign network access
+  char * buf = cs__wcstoutf(wbuf);
+  char * gbuf = guardpath(buf, 4);
+  //printf("win_open buf %s grd %s\n", buf, gbuf);
+  free(buf);
+  if (!gbuf)
+    return;
+
+  {
+#ifdef pathname_conversion_here
+#warning now deprecated; handled via guardpath
     // Need to convert POSIX path to Windows first
     if (support_wsl) {
-      // First, we need to replicate some of the handling of relative paths 
-      // as implemented in child_conv_path,
-      // because the dewsl functionality would actually go in between 
-      // the workflow of child_conv_path.
-      // We cannot determine the WSL foreground process and its 
-      // current directory, so we can only consider the working directory 
-      // explicitly communicated via the OSC 7 escape sequence here.
-      if (*wpath != '/' && wcsncmp(wpath, W("~/"), 2) != 0) {
+      // First, we need to replicate some of the handling of relative paths
+      // as implemented in child_conv_path, because the dewsl functionality
+      // would actually go in between the workflow of child_conv_path.
+      // We cannot determine the WSL foreground process and its current
+      // directory, so we can only consider the working directory explicitly
+      // communicated via the OSC 7 escape sequence here.
+      if (*wbuf != '/' && wcsncmp(wbuf, W("~/"), 2)) {
         if (child_dir && *child_dir) {
           wchar * cd = cs__mbstowcs(child_dir);
-          cd = renewn(cd, wcslen(cd) + wcslen(wpath) + 2);
+          cd = renewn(cd, wcslen(cd) + wl + 2);
           cd[wcslen(cd)] = '/';
-          wcscpy(&cd[wcslen(cd) + 1], wpath);
-          delete(wpath);
-          wpath = cd;
+          wcscpy(&cd[wcslen(cd) + 1], wbuf);
+          delete(wbuf);
+          wbuf = cd;
         }
       }
 
-      wpath = dewsl((wchar *)wpath);
+      wbuf = dewsl(wbuf);
     }
-    wstring conv_wpath = child_conv_path(wpath, adjust_dir);
-#ifdef debug_wsl
-    printf("open <%ls> <%ls>\n", wpath, conv_wpath);
+    wstring conv_wpath = child_conv_path(wbuf, adjust_dir);
+#else
+    (void)adjust_dir;
+    wchar *conv_wpath = path_posix_to_win_w(gbuf);
+    free(gbuf);
 #endif
-    delete(wpath);
+#ifdef debug_wslpath
+    printf("win_open <%ls> <%ls>\n", wbuf, conv_wpath);
+#endif
+    delete(wbuf);
     if (conv_wpath)
       shell_exec(conv_wpath); // frees conv_wpath
     else
@@ -794,6 +1020,9 @@ win_copy_as(const wchar *data, cattr *cattrs, int len, char what)
       char * textbuf = lockc;
       textbuf += 24;
       int buflen = WideCharToMultiByte(cs_get_codepage(), 0, data, len, textbuf, lenc, null, null);
+      // strip terminating NUL (#1132)
+      if (!textbuf[buflen - 1])
+        buflen--;
 
       struct {
         u_int64_t cb_sec;
@@ -866,7 +1095,7 @@ buf_path(wchar * wfn, bool convert, bool quote)
       buf_add('\\');
     char *p = fn;
     if (support_wsl) {
-#ifdef debug_wsl
+#ifdef debug_wslpath
       printf("paste <%s>\n", p);
 #endif
       char * wslp = wslpath(p);
